% Code for Generalized HKL Multi-task (gHKL-MT) 
% by Pratik Jawanpuria, J. Saketha Nath and Ganesh Ramakrishnan (IIT Bombay)

% This code is built upon the code of HKL (Francis Bach: Exploring Large Feature
% Spaces with Hierarchical Multiple Kernel Learning. Advances in Neural
% Information Processing Systems (NIPS), 2008) available at
% http://www.di.ens.fr/~fbach/hkl/index.html
% Some of the functions are used as it is from above and some of them are
% discarded/modified.

% In order to use gHKL-MT for REL application, you have to just enter the relevant {0,1} feature matrix.
% Please refer to our paper/technical report in order to know the details
% of converting a real and/or nominal valued feature matrix to {0,1}
% feature matrix

% For just running gHKL-MT, above conversion is not required.

% The current version of the code is specifically engineered for multi-class/multi-label classification setting.
% For instance, it assumes that the input feature matrix (X) is same across all the labels. 
% However, the more generic multi-task setting may also be learnt by modifying it slightly. 
% In multi-task learning setting, the tasks need not have the same input data. 

% This code is licensed under GPL licence and it is free for non-commercial use. 


clear all
clc

% load('multiclassDatasetExample.mat')%uncomment it for a trial run

%%%%%%PARAMETERS%%%%%%%%
weights = [1 .1 2];% 1st is dummy, second is the weight of TOP node, third is the parameter 'a' such that d_v=a^|v|
solver = 'smo_rel';%ONLY smo_rel is available in the current version of the program.
rhonorm = 1.1;%(1<rhonorm<=2)
loss = 'hinge';
kernel = 'polynomial';%for REL application use 'polynomial'
mingap = 1e-2; % relative duality gap for active set algorithm
maxactive = 10000;% maximum size of the active set
memory_cache = 1*1e9;% available memory in bytes (default=1e9(this has a significant impact on performance, i.e., do use your memory!))
high = 3;
low = -1;
evaluationType = 2;% 1-multi label classification problem, 2-multi class classification problem, 0-binary classification problem

% Caution: The code is mainly intended for mutli-label/class classification problem, though it also works for binary classification.
% It is advisable to use the (more efficient) gHKL code is for binary classification. 
% In case any issue arises for binary classification with this code, please e-mail to: pratik.iitb@gmail.com 

%%%%%%PARAMETERS%%%%%%%%

% LOAD DATA FILE containing feature matrix X
% X: n x d matrix where n is number of data points and d is dimension.
% Y: n x T matrix where T is the number of classes/labels. Each column is a {-1,1}^n vector. 
% For binary classification, T=1 since just one column vector can represent the two classes. 

% For REL, X should be a {0,1} matrix

d = size(X,2);

% default value of conflictMatrix
conflictMatrix = zeros(d,d);

% Concept of Conflict Matrix is added to make the program run faster in
% case we have any prior knowledge about the features of the given dataset.
% It should be changed ONLY when code is run for Rule Ensemble Learning. If
% code is run only for Generalized HKL (i.e. not REL) it SHOULD be a zero matrix.

% Conflict matrix captures the information whether the basic propositions (feature_i==1) &
% (feature_j==1) can coexist in a rule. If yes, the conflictMatrix(i,j) = conflictMatrix(i,j) = 0;
% Else conflictMatrix(i,j) = conflictMatrix(i,j) = 1;
% In case of no prior knowledge, conflictMatrix should be a zero matrix
% which means all basic propositions can co-occur with each other.
% Example of a case where prior knowledge is present is when features are
% of complementary type: A, (NOT A). We know for sure that basic propositions (A==1) and (NOT A ==1)
% cannot coexist in a rule. Code ran approximately 20-30% faster when such prior
% knowledge was available.


range = high - low + 1;
lambdas = logspace(high,low,range);% 1/C, C: misclassification penalt

% Prepare the train-test split - XTrain, YTrain, XTest, YTest

% format of xTrain - nTrain x d vector where nTrain is the number of training examples
% numTrain : a T x 1 vector where numTrain(i) = nTrain.
% format of YTrain - (nTrain*T) x 1 vector.
% Indices (sum(numTrain(1:(i-1)))+1):(sum(numTrain(1:i))) in YTrain should correspond to the y-vector ({-1,1}^numTrain(i)) of the i-th class/label.
% Similar procedure for test data: xTest YTest and (numTest)

[outputsErrSpRho,modelErrSpRho,accuraciesErrSpRho] = ghkl_ml(xTrain,YTrain,lambdas,loss,kernel,weights,numTrain,'Xtest',xTest,...
    'Ytest',YTest,'numTest',numTest,'display',0,'maxactive',maxactive,'memory_cache',memory_cache, 'rhonorm',rhonorm,'conflictMatrix',conflictMatrix,...
    'solver',solver,'mingap',mingap,'evaluationType',evaluationType);

% In order to see the rules (for REL application) please use the following command:
% graphModel = getGraph(outputsRho{1}.hull(outputsRho{1}.theta > 0,:))
% graphModel will be a r x 1 cell where r is the number of rules
% each row of the cell will contain one rule (indices will represent the
% feature indices of feature matrix). For example: if graphModel = {[1 2 4], [3 4 5]};
% it means the rules are: 
% R1: feature_1 AND feature_2 AND feature_4
% R2: feature_3 AND feature_4 AND feature_5
% Whether these rules favour +ve or -ve class can be see by calculating the
% weight vectors